# -*- coding: utf-8 -*-
"""
License: BSD

(c) 2009      ::: www.CodeResort.com - BV Network AS (simon-code@bvnetwork.no)
"""

import re
import datetime
from types import GeneratorType

from trac.core import *
from trac.perm import PermissionError
from trac.util.datefmt import utc
from trac.util.text import to_unicode

from tracrpc.api import IRPCProtocol, XMLRPCSystem, Binary
from tracrpc.util import exception_to_unicode, empty, prepare_docs

__all__ = ['JsonRpcProtocol']

try:
    import json
    if not (hasattr(json, 'JSONEncoder') \
            and hasattr(json, 'JSONDecoder')):
        raise AttributeError("Incorrect JSON library found.")
except (ImportError, AttributeError):
    import simplejson as json

class TracRpcJSONEncoder(json.JSONEncoder):
    """ Extending the JSON encoder to support some additional types:
    1. datetime.datetime => {'__jsonclass__': ["datetime", "<rfc3339str>"]}
    2. tracrpc.api.Binary => {'__jsonclass__': ["binary", "<base64str>"]}
    3. empty => '' """

    def default(self, obj):
        if isinstance(obj, datetime.datetime):
            # http://www.ietf.org/rfc/rfc3339.txt
            return {'__jsonclass__': ["datetime",
                            obj.strftime('%Y-%m-%dT%H:%M:%S')]}
        elif isinstance(obj, Binary):
            return {'__jsonclass__': ["binary",
                            obj.data.encode("base64")]}
        elif obj is empty:
            return ''
        else:
            return json.JSONEncoder(self, obj)

class TracRpcJSONDecoder(json.JSONDecoder):
    """ Extending the JSON decoder to support some additional types:
    1. {'__jsonclass__': ["datetime", "<rfc3339str>"]} => datetime.datetime
    2. {'__jsonclass__': ["binary", "<base64str>"]} => tracrpc.api.Binary """

    dt = re.compile(
        '^(\d{4})-(\d{2})-(\d{2})T(\d{2}):(\d{2}):(\d{2})(?:\.(\d{1,}))?')

    def _normalize(self, obj):
        """ Helper to traverse JSON decoded object for custom types. """
        if isinstance(obj, tuple):
            return tuple(self._normalize(item) for item in obj)
        elif isinstance(obj, list):
            return [self._normalize(item) for item in obj]
        elif isinstance(obj, dict):
            if obj.keys() == ['__jsonclass__']:
                kind, val = obj['__jsonclass__']
                if kind == 'datetime':
                    dt = self.dt.match(val)
                    if not dt:
                        raise Exception(
                                "Invalid datetime string (%s)" % val)
                    dt = tuple([int(i) for i in dt.groups() if i])
                    kw_args = {'tzinfo': utc}
                    return datetime.datetime(*dt, **kw_args)
                elif kind == 'binary':
                    try:
                        bin = val.decode("base64")
                        return Binary(bin)
                    except:
                        raise Exception("Invalid base64 string")
                else:
                    raise Exception("Unknown __jsonclass__: %s" % kind)
            else:
                return dict(self._normalize(obj.items()))
        elif isinstance(obj, basestring):
            return to_unicode(obj)
        else:
            return obj

    def decode(self, obj, *args, **kwargs):
        obj = json.JSONDecoder.decode(self, obj, *args, **kwargs)
        return self._normalize(obj)

class JsonRpcProtocol(Component):
    r"""
    Example `POST` request using `curl` with `Content-Type` header
    and body:
 
    {{{
    user: ~ > cat body.json
    {"params": ["WikiStart"], "method": "wiki.getPage", "id": 123}
    user: ~ > curl -H "Content-Type: application/json" --data @body.json ${req.abs_href.rpc()}
    {"id": 123, "error": null, "result": "= Welcome to....
    }}}
    
    Implementation details:
    
      * JSON-RPC has no formalized type system, so a class-hint system is used
        for input and output of non-standard types:
        * `{"__jsonclass__": ["datetime", "YYYY-MM-DDTHH:MM:SS"]} => DateTime (UTC)`
        * `{"__jsonclass__": ["binary", "<base64-encoded>"]} => Binary`
      * `"id"` is optional, and any marker value received with a
        request is returned with the response.
    """

    implements(IRPCProtocol)

    # IRPCProtocol methods

    def rpc_info(self):
        return ('JSON-RPC', prepare_docs(self.__doc__))

    def rpc_match(self):
        yield('rpc', 'application/json')
        # Legacy path - provided for backwards compatibility:
        yield ('jsonrpc', 'application/json')

    def rpc_process(self, req, content_type):
        """ Handles JSON-RPC requests """
        if not json:
            self.log.debug("RPC(json) call ignored (not available).")
            self._send_response(req, "Error: JSON-RPC not available.\n",
                                content_type)
        try:
            data = json.load(req, cls=TracRpcJSONDecoder)
        except Exception, e:
            # Abort with exception - no data can be read
            self.log.error("RPC(json) decode error %s" % \
                    exception_to_unicode(e, traceback=True))
            response = json.dumps(self._json_error(e, -32700),
                                    cls=TracRpcJSONEncoder)
            self._send_response(req, response + '\n', content_type)
            return
        self.log.debug("RPC(json) call by '%s': %s" % (req.authname, data))
        args = data.get('params') or []
        r_id = data.get('id', None)
        method = data.get('method', '')
        try:
            req.perm.require('XML_RPC') # Need at least XML_RPC
            if method == 'system.multicall': # Custom multicall
                results = []
                for mc in args:
                    results.append(self._json_call(req, mc.get('method', ''),
                        mc.get('params') or [], mc.get('id') or r_id))
                response = {'result': results, 'error': None, 'id': r_id}
            else:
                response = self._json_call(req, method, args, r_id)
            try: # JSON encoding
                self.log.debug("RPC(json) result: %s" % repr(response))
                response = json.dumps(response, cls=TracRpcJSONEncoder)
            except Exception, e:
                response = json.dumps(self._json_error(e, r_id=r_id),
                                        cls=TracRpcJSONEncoder)
        except PermissionError, e:
            response = json.dumps(self._json_error(e, -32600, r_id=r_id),
                cls=TracRpcJSONEncoder)
        except Exception, e:
            self.log.error("RPC(json) error %s" % exception_to_unicode(e,
                                                    traceback=True))
            response = json.dumps(self._json_error(e), cls=TracRpcJSONEncoder)
        self.log.debug("RPC(json) encoded result: %s" % response)
        self._send_response(req, response + '\n', content_type)

    # Internal methods

    def _send_response(self, req, response, content_type='application/json'):
        response = to_unicode(response).encode("utf-8")
        req.send_response(200)
        req.send_header('Content-Type', content_type)
        req.send_header('Content-Length', len(response))
        req.end_headers()
        req.write(response)

    def _json_call(self, req, method, args, r_id=None):
        """ Call method and create response dictionary. """
        try:
            result = (XMLRPCSystem(self.env).get_method(method)(req, args))[0]
            if isinstance(result, GeneratorType):
                result = list(result)
            return {'result': result, 'error': None, 'id': r_id}
        except Exception, e:
            return self._json_error(e, r_id=r_id)

    def _json_error(self, e, c=-32603, r_id=None):
        """ Makes a response dictionary that is an error. """
        return {'result': None, 'id': r_id, 'error': {
                'name': 'JSONRPCError', 'code': c, 'message': to_unicode(e)}}

